package org.anjocaido.groupmanager.dataholder;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.anjocaido.groupmanager.data.Group;

/**
 * This container holds all Groups loaded from the relevant groupsFile.
 *
 * @author ElgarL
 *
 */
public class GroupsDataHolder {
    private WorldDataHolder dataSource;
    private Group defaultGroup = null;
    private File groupsFile;
    private boolean haveGroupsChanged = false;
    private long timeStampGroups = 0;
    /**
     * The actual groups holder
     */
    private final Map<String, Group> groups = Collections.synchronizedMap(new HashMap<String, Group>());

    /**
     * Constructor
     */
    protected GroupsDataHolder() {
    }

    public WorldDataHolder getDataSource() {
        return this.dataSource;
    }

    /**
     * @return the defaultGroup
     */
    public Group getDefaultGroup() {
        return defaultGroup;
    }

    /**
     * Note: Iteration over this object has to be synchronized!
     *
     * @return the groups
     */
    public Map<String, Group> getGroups() {
        return groups;
    }

    /**
     * @return the groupsFile
     */
    public File getGroupsFile() {
        return groupsFile;
    }

    /**
     * @return the timeStampGroups
     */
    public long getTimeStampGroups() {
        return timeStampGroups;
    }

    /**
     * @return the haveGroupsChanged
     */
    public boolean HaveGroupsChanged() {
        return haveGroupsChanged;
    }

    /**
     * Resets the Groups
     */
    public void resetGroups() {
        this.groups.clear();
    }

    public void setDataSource(final WorldDataHolder dataSource) {
        this.dataSource = dataSource;
        // push this data source to the users, so they pull the correct groups data.
        synchronized (groups) {
            for (final Group group : groups.values())
                group.setDataSource(this.dataSource);
        }
    }

    /**
     * @param defaultGroup
     *            the defaultGroup to set
     */
    public void setDefaultGroup(final Group defaultGroup) {
        this.defaultGroup = defaultGroup;
    }

    /**
     * @param haveGroupsChanged
     *            the haveGroupsChanged to set
     */
    public void setGroupsChanged(final boolean haveGroupsChanged) {
        this.haveGroupsChanged = haveGroupsChanged;
    }

    /**
     * @param groupsFile
     *            the groupsFile to set
     */
    public void setGroupsFile(final File groupsFile) {
        this.groupsFile = groupsFile;
    }

    /**
     * @param timeStampGroups
     *            the timeStampGroups to set
     */
    public void setTimeStampGroups(final long timeStampGroups) {
        this.timeStampGroups = timeStampGroups;
    }
}
